import os
import polars as pl
import sys
# noinspection PyUnresolvedReferences,PyProtectedMember
from PheTK import _queries, _utils


class Phecode:
    """
    Class Phecode extract ICD codes and map ICD codes to phecodes version 1.2 and X.
    Currently, supports ICD code extraction for All of Us OMOP data.
    For other databases, user is expected to provide an ICD code table for all participants in cohort of interest.
    """

    def __init__(self, platform="aou", icd_df_path=None):
        """
        Instantiate based on parameter db
        :param platform: supports:
            "aou": All of Us OMOP database
            "custom": other databases; icd_df must be not None if db = "custom"
        :param icd_df_path: path to ICD table csv file; required columns are "person_id", "ICD", and "vocabulary_id";
            "vocabulary_id" values should be "ICD9CM" or "ICD10CM"
        """
        self.platform = platform

        if platform == "aou":
            self.cdr = os.getenv("WORKSPACE_CDR")
            self.icd_query = _queries.phecode_icd_query(self.cdr)
            print("\033[1mStart querying ICD codes...")
            self.icd_events = _utils.polars_gbq(self.icd_query)

        elif platform == "custom":
            if icd_df_path is not None:
                print("\033[1mLoading user's ICD data from file...")
                self.icd_events = pl.read_csv(icd_df_path,
                                              dtypes={"ICD": str})
            else:
                print("icd_df_path is required for custom platform.")
                sys.exit(0)
        else:
            print("Invalid platform. Parameter platform only accepts \"aou\" (All of Us) or \"custom\".")
            sys.exit(0)

        # add flag column if not exist
        if "flag" not in self.icd_events.columns:
            self.icd_events = self.icd_events.with_columns(
                pl.when((pl.col("vocabulary_id") == "ICD9") |
                        (pl.col("vocabulary_id") == "ICD9CM"))
                .then(9)
                .when((pl.col("vocabulary_id") == "ICD10") |
                      (pl.col("vocabulary_id") == "ICD10CM"))
                .then(10)
                .otherwise(0)
                .alias("flag")
                .cast(pl.Int8)
            )
        else:
            self.icd_events = self.icd_events.with_columns(pl.col("flag").cast(pl.Int8))

        print("\033[1mDone!")

    def count_phecode(self, phecode_version="X", icd_version="US",
                      phecode_map_file_path=None, output_file_name=None):
        """
        Generate phecode counts from ICD counts
        :param phecode_version: defaults to "X"; other option is "1.2"
        :param icd_version: defaults to "US"; other option are "WHO" and "custom";
                            if "custom", user need to provide phecode_map_path
        :param phecode_map_file_path: path to custom phecode map table
        :param output_file_name: user specified output file name
        :return: phecode counts polars dataframe
        """        
        # load phecode mapping file by version or by custom path
        phecode_df = _utils.get_phecode_mapping_table(
            phecode_version=phecode_version,
            icd_version=icd_version,
            phecode_map_file_path=phecode_map_file_path,
            keep_all_columns=False
        )

        # make a copy of self.icd_events
        icd_events = self.icd_events.clone()

        # keep only necessary columns
        icd_events = icd_events[["person_id", "ICD", "flag"]]

        print()
        print(f"\033[1mMapping ICD codes to phecode {phecode_version}...")
        if phecode_version == "X":
            phecode_counts = icd_events.join(phecode_df,
                                             how="inner",
                                             on=["ICD", "flag"])
        elif phecode_version == "1.2":
            phecode_counts = icd_events.join(phecode_df,
                                             how="inner",
                                             on=["ICD", "flag"])
            phecode_counts = phecode_counts.rename({"phecode_unrolled": "phecode"})
        else:
            phecode_counts = pl.DataFrame()
            
        if not phecode_counts.is_empty():
            phecode_counts = phecode_counts.group_by(["person_id", "phecode"]).len().rename({"len": "count"})

        # report result
        if not phecode_counts.is_empty():
            if output_file_name is None:
                file_name = "{0}_{1}_phecode{2}_counts.csv".format(self.platform, icd_version,
                                                                   phecode_version.upper().replace(".", ""))
            else:
                file_name = output_file_name
            phecode_counts.write_csv(file_name)
            print(f"\033[1mSuccessfully generated phecode {phecode_version} counts for cohort participants!\n"
                  f"\033[1mSaved to {file_name}!\033[0m")
            print()

        else:
            print("\033[1mNo phecode count generated. Check your input data.\033[0m")
            print()



from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from tqdm import tqdm
import argparse
import copy
import numpy as np
import os
import pandas as pd
import polars as pl
import statsmodels
import statsmodels.api as sm
import sys
import warnings
# noinspection PyUnresolvedReferences,PyProtectedMember
from PheTK import _utils


class PheWAS:

    def __init__(self,
                 phecode_version,
                 phecode_count_csv_path,
                 cohort_csv_path,
                 sex_at_birth_col,
                 covariate_cols,
                 independent_variable_of_interest,
                 male_as_one=True,
                 icd_version="US",
                 phecode_map_file_path=None,
                 phecode_to_process="all",
                 min_cases=50,
                 min_phecode_count=2,
                 use_exclusion=False,
                 output_file_name=None,
                 verbose=False,
                 suppress_warnings=True):
        """
        :param phecode_version: accepts "1.2" or "X"
        :param phecode_count_csv_path: path to phecode count of relevant participants at minimum
        :param cohort_csv_path: path to cohort data with covariates of interest
        :param sex_at_birth_col: gender/sex column of interest, by default, male = 1, female = 0;
                                 can use male_as_one parameter to specify the value assigned to male
        :param covariate_cols: name of covariate columns; excluding independent var of interest
        :param independent_variable_of_interest: independent variable of interest column name
        :param male_as_one: defaults to True; if True, male=1 and female=0; if False, male=0 and female=1;
                            use this to match how males and females are coded in sex_at_birth column;
        :param icd_version: defaults to "US"; other option are "WHO" and "custom";
                            if "custom", user need to provide phecode_map_path
        :param phecode_map_file_path: path to custom phecode map table
        :param phecode_to_process: defaults to "all"; otherwise, a list of phecodes must be provided
        :param min_cases: defaults to 50; minimum number of cases for each phecode to be considered for PheWAS
        :param min_phecode_count: defaults to 2; minimum number of phecode count to qualify as case for PheWAS
        :param use_exclusion: defaults to True for phecode 1.2; always False for phecode X;
                              whether to use additional exclusion range in control for PheWAS
        :param output_file_name: if None, defaults to "phewas_{timestamp}.csv"
        :param verbose: defaults to False; if True, print brief result of each phecode run
        :param suppress_warnings: defaults to True;
                                  if True, ignore common exception warnings such as ConvergenceWarnings, etc.
        """
        print("~~~~~~~~~~~~~~~~~~~~~~~~    Creating PheWAS Object    ~~~~~~~~~~~~~~~~~~~~~~~~~~")

        # load phecode mapping file by version or by custom path
        self.phecode_df = _utils.get_phecode_mapping_table(
            phecode_version=phecode_version,
            icd_version=icd_version,
            phecode_map_file_path=phecode_map_file_path,
            keep_all_columns=True
        )

        # load phecode counts data for all participants
        # noinspection PyTypeChecker
        self.phecode_counts = pl.read_csv(phecode_count_csv_path,
                                          dtypes={"phecode": str})

        # load covariate data
        # make sure person_id in covariate data has the same type as person_id in phecode count
        self.covariate_df = pl.read_csv(cohort_csv_path)
        self.covariate_df = self.covariate_df.drop_nulls()

        # basic attributes from instantiation
        self.sex_at_birth_col = sex_at_birth_col
        if isinstance(covariate_cols, str):
            self.covariate_cols = [copy.deepcopy(covariate_cols)]
        elif isinstance(covariate_cols, list):
            self.covariate_cols = copy.deepcopy(covariate_cols)
        self.independent_variable_of_interest = independent_variable_of_interest
        self.verbose = verbose
        self.min_cases = min_cases
        self.min_phecode_count = min_phecode_count
        self.suppress_warnings = suppress_warnings

        # assign 1 & 0 to male & female based on male_as_one parameter
        if male_as_one:
            self.male_value = 1
            self.female_value = 0
        else:
            self.male_value = 0
            self.female_value = 1

        # exclusion:
        # - phecode 1.2: user can choose to use exclusion or not
        # - phecode X: exclusion is removed, therefore this parameter will be False for Phecode X regardless of input
        # to prevent user error
        if phecode_version == "1.2":
            self.use_exclusion = use_exclusion
        elif phecode_version == "X":
            self.use_exclusion = False

        # check if variable_of_interest is included in covariate_cols
        self.variable_of_interest_in_covariates = False
        if self.independent_variable_of_interest in self.covariate_cols:
            self.variable_of_interest_in_covariates = True
            self.covariate_cols.remove(self.independent_variable_of_interest)
            print()
            print(f"Note: \"{self.independent_variable_of_interest}\" will not be used as covariate",
                  "since it was already specified as variable of interest.")
            print()

        # remove sex_at_birth column from covariates if it is included, just for processing purpose
        self.sex_as_covariate = False
        if self.sex_at_birth_col in self.covariate_cols:
            self.sex_as_covariate = True
            self.covariate_cols.remove(self.sex_at_birth_col)

        # check for sex in data
        self.data_has_single_sex = False
        self.gender_specific_var_cols = [self.independent_variable_of_interest] + self.covariate_cols
        self.sex_values = self.covariate_df[sex_at_birth_col].unique().to_list()
        # when cohort only has 1 sex with 0 or 1 as value
        if (len(self.sex_values) == 1) and ((0 in self.sex_values) or (1 in self.sex_values)):
            self.data_has_single_sex = True
            self.single_sex_value = self.covariate_df[sex_at_birth_col].unique().to_list()[0]
            self.var_cols = [self.independent_variable_of_interest] + self.covariate_cols
            # when cohort only has 1 sex, and sex was chosen as a covariate
            if self.sex_as_covariate:
                print()
                print(f"Note: \"{self.sex_at_birth_col}\" will not be used as covariate",
                      "since there is only one sex in data.")
                print()
            # when cohort only has 1 sex, and variable_of_interest is also sex_at_birth column
            if self.independent_variable_of_interest == self.sex_at_birth_col:
                print()
                print(f"Warning: Cannot use \"{self.sex_at_birth_col}\" as variable of interest in single sex cohorts.")
                print()
                sys.exit(0)
        # when cohort has 2 sexes
        elif len(self.sex_values) == 2 and ((0 in self.sex_values) and (1 in self.sex_values)):
            if self.independent_variable_of_interest == self.sex_at_birth_col:
                self.var_cols = self.covariate_cols + [self.sex_at_birth_col]
            else:
                if not self.sex_as_covariate:
                    print()
                    print("Warning: Data has both sexes but user did not specify sex as a covariate.")
                    print("         Running PheWAS without sex as a covariate.")
                    print()
                self.var_cols = [self.independent_variable_of_interest] + self.covariate_cols + [self.sex_at_birth_col]
        # all other cases where sex_at_birth_column not coded probably
        else:
            print(f"Warning: Please check column {self.sex_at_birth_col}.")
            print("          This column should have upto 2 unique values, 0 for female and 1 for male.")
            sys.exit(0)

        # check for string type variables among covariates
        if pl.Utf8 in self.covariate_df[self.var_cols].schema.values():
            str_cols = [k for k, v in self.covariate_df.schema.items() if v is pl.Utf8]
            print(f"Column(s) {str_cols} contain(s) string type. Only numerical types are accepted.")
            sys.exit(0)

        # keep only relevant columns in covariate_df
        cols_to_keep = list(set(["person_id"] + self.var_cols))
        self.covariate_df = self.covariate_df[cols_to_keep]
        self.covariate_df = self.covariate_df.drop_nulls()
        self.cohort_size = self.covariate_df.n_unique()

        # update phecode_counts to only participants of interest
        self.cohort_ids = self.covariate_df["person_id"].unique().to_list()
        self.phecode_counts = self.phecode_counts.filter(pl.col("person_id").is_in(self.cohort_ids))
        if phecode_to_process == "all":
            self.phecode_list = self.phecode_counts["phecode"].unique().to_list()
        else:
            if isinstance(phecode_to_process, str):
                phecode_to_process = [phecode_to_process]
            self.phecode_list = phecode_to_process

        # attributes for reporting PheWAS results
        self._phecode_summary_statistics = None
        self._cases = None
        self._controls = None
        self.results = None
        self.not_tested_count = 0
        self.tested_count = 0
        self.bonferroni = None
        self.phecodes_above_bonferroni = None
        self.above_bonferroni_count = None

        # for saving results
        if output_file_name is not None:
            if ".csv" in output_file_name:
                output_file_name = output_file_name.replace(".csv", "")
            self.output_file_name = output_file_name + ".csv"
        else:
            self._timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            self.output_file_name = f"phewas_{self._timestamp}.csv"

    @staticmethod
    def _to_polars(df):
        """
        Check and convert pandas dataframe object to polars dataframe, if applicable
        :param df: dataframe object
        :return: polars dataframe
        """
        if isinstance(df, pd.DataFrame):
            polars_df = pl.from_pandas(df)
        else:
            polars_df = df
        return polars_df

    def _exclude_range(self, phecode, phecode_df=None):
        """
        Process text data in exclude_range column; exclusively for phecodeX
        :param phecode: phecode of interest
        :return: processed exclude_range, either None or a valid list of phecode(s)
        """
        if phecode_df is None:
            phecode_df = self.phecode_df.clone()

        # not all phecode has exclude_range
        # exclude_range can be single code (e.g., "777"), single range (e.g., "777-780"),
        # or multiple ranges/codes (e.g., "750-777,586.2")
        phecodes_without_exclude_range = phecode_df.filter(
            pl.col("exclude_range").is_null()
        )["phecode"].unique().to_list()
        if phecode in phecodes_without_exclude_range:
            exclude_range = []
        else:
            # get exclude_range value of phecode
            ex_val = phecode_df.filter(pl.col("phecode") == phecode)["exclude_range"].unique().to_list()[0]

            # split multiple codes/ranges
            comma_split = ex_val.split(",")
            if len(comma_split) == 1:
                exclude_range = comma_split
            else:
                exclude_range = []
                for item in comma_split:
                    # process range in text form
                    if "-" in item:
                        first_code = item.split("-")[0]
                        last_code = item.split("-")[1]
                        dash_range = [str(i) for i in range(int(first_code), int(last_code))] + [last_code]
                        exclude_range = exclude_range + dash_range
                    else:
                        exclude_range = exclude_range + [item]

        return exclude_range

    def _case_control_prep(self, phecode,
                           phecode_counts=None, covariate_df=None, phecode_df=None,
                           var_cols=None, gender_specific_var_cols=None, keep_ids=False):
        """
        :param phecode: phecode of interest
        :param phecode_counts: phecode counts table for cohort
        :param covariate_df: covariate table for cohort
        :param phecode_df: phecode mapping table
        :param var_cols: variable columns in general case
        :param gender_specific_var_cols: variable columns in gender-specific case
        :param keep_ids: if True, keep phecode person_id column, used when person_id is needed
        :return: cases, controls and analysis_var_cols
        """
        if phecode_counts is None:
            phecode_counts = self.phecode_counts.clone()
        if covariate_df is None:
            covariate_df = self.covariate_df.clone()
        if phecode_df is None:
            phecode_df = self.phecode_df.clone()
        if var_cols is None:
            var_cols = copy.deepcopy(self.var_cols)
        if gender_specific_var_cols is None:
            gender_specific_var_cols = copy.deepcopy(self.gender_specific_var_cols)

        # SEX RESTRICTION
        filtered_phecode_df = phecode_df.filter(pl.col("phecode") == phecode)
        if len(filtered_phecode_df["sex"].unique().to_list()) > 0:
            sex_restriction = filtered_phecode_df["sex"].unique().to_list()[0]
        else:
            return pl.DataFrame(), pl.DataFrame(), []

        # ANALYSIS VAR COLS
        if sex_restriction == "Both" and self.sex_as_covariate:
            analysis_var_cols = var_cols
        else:
            analysis_var_cols = gender_specific_var_cols

        # FILTER COVARIATE DATA BY PHECODE SEX RESTRICTION
        if not self.data_has_single_sex:
            if sex_restriction == "Male":
                covariate_df = covariate_df.filter(pl.col(self.sex_at_birth_col) == self.male_value)
            elif sex_restriction == "Female":
                covariate_df = covariate_df.filter(pl.col(self.sex_at_birth_col) == self.female_value)
        else:
            if (
                (sex_restriction == "Male" and self.male_value not in self.sex_values)
                or (sex_restriction == "Male") and (self.male_value not in self.sex_values)
            ):
                return pl.DataFrame(), pl.DataFrame(), []

        # GENERATE CASES & CONTROLS
        if len(covariate_df) > 0:
            # CASE
            # participants with at least <min_phecode_count> phecodes
            case_ids = phecode_counts.filter(
                (pl.col("phecode") == phecode) & (pl.col("count") >= self.min_phecode_count)
            )["person_id"].unique().to_list()
            cases = covariate_df.filter(pl.col("person_id").is_in(case_ids))

            # CONTROLS
            # phecode exclusions
            if self.use_exclusion:
                exclude_range = [phecode] + self._exclude_range(phecode, phecode_df=phecode_df)
            else:
                exclude_range = [phecode]
            # get participants ids to exclude and filter covariate df
            exclude_ids = phecode_counts.filter(
                pl.col("phecode").is_in(exclude_range)
            )["person_id"].unique().to_list()
            controls = covariate_df.filter(~(pl.col("person_id").is_in(exclude_ids)))

            # DUPLICATE CHECK
            # drop duplicates
            duplicate_check_cols = ["person_id"] + analysis_var_cols
            cases = cases.unique(subset=duplicate_check_cols)
            controls = controls.unique(subset=duplicate_check_cols)

            if not keep_ids:
                # KEEP ONLY REQUIRED COLUMNS
                cases = cases[analysis_var_cols]
                controls = controls[analysis_var_cols]
            else:
                cases = cases[duplicate_check_cols]
                controls = controls[duplicate_check_cols]

            return cases, controls, analysis_var_cols

        else:
            return pl.DataFrame(), pl.DataFrame(), []

    def get_phecode_data(self, phecode):
        cases, controls, analysis_var_cols = self._case_control_prep(
            phecode=phecode,
            keep_ids=True
        )

        if len(cases) >= 0 or len(controls) >= 0:
            cases = cases.with_columns(pl.lit(True).alias("is_phecode_case"))
            controls = controls.with_columns(pl.lit(False).alias("is_phecode_case"))
            phecode_data = cases.vstack(controls)

            return phecode_data

        else:
            print(f"No phecode data for {phecode}")

    @staticmethod
    def _result_prep(result, var_of_interest_index):
        """
        Process result from statsmodels
        :param result: logistic regression result
        :param var_of_interest_index: index of variable of interest
        :return: dataframe with key statistics
        """
        results_as_html = result.summary().tables[0].as_html()
        converged = pd.read_html(results_as_html)[0].iloc[5, 1]
        results_as_html = result.summary().tables[1].as_html()
        res = pd.read_html(results_as_html, header=0, index_col=0)[0]

        p_value = result.pvalues[var_of_interest_index]
        neg_log_p_value = -np.log10(p_value)
        beta = result.params[var_of_interest_index]
        conf_int_1 = res.iloc[var_of_interest_index]['[0.025']
        conf_int_2 = res.iloc[var_of_interest_index]['0.975]']
        odds_ratio = np.exp(beta)
        log10_odds_ratio = np.log10(odds_ratio)

        return {"p_value": p_value,
                "neg_log_p_value": neg_log_p_value,
                "beta": beta,
                "conf_int_1": conf_int_1,
                "conf_int_2": conf_int_2,
                "odds_ratio": odds_ratio,
                "log10_odds_ratio": log10_odds_ratio,
                "converged": converged}

    def _logistic_regression(self, phecode,
                             phecode_counts=None, covariate_df=None,
                             var_cols=None, gender_specific_var_cols=None):
        """
        Logistic regression of single phecode
        :param phecode: phecode of interest
        :param phecode_counts: phecode counts table for cohort
        :param covariate_df: covariate table for cohort
        :param var_cols: variable columns in general case
        :param gender_specific_var_cols: variable columns in gender-specific case
        :return: result_dict object
        """

        if phecode_counts is None:
            phecode_counts = self.phecode_counts.clone()
        if covariate_df is None:
            covariate_df = self.covariate_df.clone()
        if var_cols is None:
            var_cols = copy.deepcopy(self.var_cols)
        if gender_specific_var_cols is None:
            gender_specific_var_cols = copy.deepcopy(self.gender_specific_var_cols)

        cases, controls, analysis_var_cols = self._case_control_prep(phecode,
                                                                     phecode_counts=phecode_counts,
                                                                     covariate_df=covariate_df,
                                                                     var_cols=var_cols,
                                                                     gender_specific_var_cols=gender_specific_var_cols)

        # only run regression if number of cases > min_cases
        if (len(cases) >= self.min_cases) and (len(controls) > self.min_cases):

            # add case/control values
            cases = cases.with_columns(pl.Series([1] * len(cases)).alias("y"))
            controls = controls.with_columns(pl.Series([0] * len(controls)).alias("y"))

            # merge cases & controls
            regressors = cases.vstack(controls)

            # get index of variable of interest
            var_index = regressors[analysis_var_cols].columns.index(self.independent_variable_of_interest)

            # logistic regression
            if self.suppress_warnings:
                warnings.simplefilter("ignore")
            y = regressors["y"].to_numpy()
            regressors = regressors[analysis_var_cols].to_numpy()
            regressors = sm.tools.add_constant(regressors, prepend=False)
            logit = sm.Logit(y, regressors, missing="drop")

            # catch Singular matrix error
            try:
                result = logit.fit(disp=False)
            except (np.linalg.linalg.LinAlgError, statsmodels.tools.sm_exceptions.PerfectSeparationError) as err:
                if "Singular matrix" in str(err) or "Perfect separation" in str(err):
                    if self.verbose:
                        print(f"Phecode {phecode} ({len(cases)} cases/{len(controls)} controls):", str(err), "\n")
                    pass
                else:
                    raise
                result = None

            if result is not None:
                # process result
                base_dict = {"phecode": phecode,
                             "cases": len(cases),
                             "controls": len(controls)}
                stats_dict = self._result_prep(result=result, var_of_interest_index=var_index)
                result_dict = {**base_dict, **stats_dict}  # python 3.5 or later
                # result_dict = base_dict | stats_dict  # python 3.9 or later

                # choose to see results on the fly
                if self.verbose:
                    print(f"Phecode {phecode} ({len(cases)} cases/{len(controls)} controls): {result_dict}\n")

                return result_dict

        else:
            if self.verbose:
                print(f"Phecode {phecode} ({len(cases)} cases/{len(controls)} controls):",
                      "Not enough cases or controls. Pass.\n")

    def run(self,
            parallelization="multithreading",
            n_threads=round(os.cpu_count()*2/3)):
        """
        Run parallel logistic regressions
        :param parallelization: defaults to "multithreading", utilizing concurrent.futures.ThreadPoolExecutor();
                                if "multiprocessing": use multiprocessing.Pool()
        :param n_threads: number of threads in multithreading
        :return: PheWAS summary statistics Polars dataframe
        """
        print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~    Running PheWAS    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")

        if parallelization == "multithreading":
            with ThreadPoolExecutor(max_workers=n_threads) as executor:
                jobs = [
                    executor.submit(
                        self._logistic_regression,
                        phecode,
                        self.phecode_counts.clone(),
                        self.covariate_df.clone(),
                        copy.deepcopy(self.var_cols),
                        copy.deepcopy(self.gender_specific_var_cols)
                    ) for phecode in self.phecode_list
                ]
                result_dicts = [job.result() for job in tqdm(as_completed(jobs), total=len(self.phecode_list))]
        else:
            return "Invalid parallelization method! Currently only supports \"multithreading\""
        result_dicts = [result for result in result_dicts if result is not None]
        if result_dicts:
            result_df = pl.from_dicts(result_dicts)
            self.results = result_df.join(self.phecode_df[["phecode", "sex",
                                                           "phecode_string",
                                                           "phecode_category"]].unique(),
                                          how="left",
                                          on="phecode").rename({"sex": "phecode_sex_restriction"})

            print("~~~~~~~~~~~~~~~~~~~~~~~~~~~    PheWAS Completed    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")

            self.tested_count = len(self.results)
            self.not_tested_count = len(self.phecode_list) - self.tested_count
            self.bonferroni = -np.log10(0.05 / self.tested_count)
            self.phecodes_above_bonferroni = self.results.filter(pl.col("neg_log_p_value") > self.bonferroni)
            self.above_bonferroni_count = len(self.phecodes_above_bonferroni)

            # save results
            self.results.write_csv(self.output_file_name)

            print("Number of participants in cohort:", self.cohort_size)
            print("Number of phecodes in cohort:", len(self.phecode_list))
            print(f"Number of phecodes having less than {self.min_cases} cases or controls:", self.not_tested_count)
            print("Number of phecodes tested:", self.tested_count)
            print(u"Suggested Bonferroni correction (-log\u2081\u2080 scale):", self.bonferroni)
            print("Number of phecodes above Bonferroni correction:", self.above_bonferroni_count)
            print()
            print("PheWAS results saved to\033[1m", self.output_file_name, "\033[0m")
        else:
            print("No analysis done. Please check your inputs.")

        print()


def main():

    # parse args
    parser = argparse.ArgumentParser(description="PheWAS analysis tool.")
    parser.add_argument("-p",
                        "--phecode_count_csv_path",
                        type=str, required=True,
                        help="Path to the phecode count csv file.")
    parser.add_argument("-c",
                        "--cohort_csv_path",
                        type=str, required=True,
                        help="Path to the cohort csv file.")
    parser.add_argument("-pv",
                        "--phecode_version",
                        type=str, required=True, choices=["1.2", 'X'],
                        help="Phecode version.")
    parser.add_argument("-cv",
                        "--covariates",
                        nargs="+",
                        type=str, required=True,
                        help="List of covariates to use in PheWAS analysis.")
    parser.add_argument("-i",
                        "--independent_variable_of_interest",
                        type=str, required=True,
                        help="Independent variable of interest.")
    parser.add_argument("-s",
                        "--sex_at_birth_col",
                        type=str, required=True,
                        help="Sex at birth column.")
    parser.add_argument("-mso",
                        "--male_as_one",
                        type=bool, required=False,
                        help="Whether male was assigned as 1 in data.")
    parser.add_argument("-pl",
                        "--phecode_to_process",
                        nargs="+",
                        type=str, required=False, default="all",
                        help="List of specific phecodes to use in PheWAS analysis.")
    parser.add_argument("-e",
                        "--use_exclusion",
                        type=bool, required=False, default=False,
                        help="Whether to use phecode exclusions. Only applicable for phecode 1.2.")
    parser.add_argument("-mc",
                        "--min_case",
                        type=int, required=False, default=50,
                        help="Minimum number of cases required to be tested.")
    parser.add_argument("-mpc",
                        "--min_phecode_count",
                        type=int, required=False, default=2,
                        help="Minimum number of phecode counts required to be considered as case.")
    parser.add_argument("-t",
                        "--threads",
                        type=int, required=False, default=round(os.cpu_count()*2/3),
                        help="Number of threads to use for parallel.")
    parser.add_argument("-o",
                        "--output_file_name",
                        type=str, required=False, default="phewas_results.csv")
    args = parser.parse_args()

    # run PheWAS
    phewas = PheWAS(phecode_version=args.phecode_version,
                    phecode_count_csv_path=args.phecode_count_csv_path,
                    cohort_csv_path=args.cohort_csv_path,
                    sex_at_birth_col=args.sex_at_birth_col,
                    male_as_one=args.male_as_one,
                    covariate_cols=args.covariates,
                    independent_variable_of_interest=args.independent_variable_of_interest,
                    phecode_to_process=args.phecode_to_process,
                    use_exclusion=args.use_exclusion,
                    min_cases=args.min_case,
                    min_phecode_count=args.min_phecode_count,
                    output_file_name=args.output_file_name)
    phewas.run(n_threads=args.threads)


if __name__ == "__main__":
    main()



def phecode_icd_query(ds):
    """
    This method is optimized for All of Us platform.

    It includes 3 queries: icd_query, v_icd_vocab_query, and final_query.
    icd_query retrieves all ICD codes from OMOP database.
    v_icd_vocab_query get the ICD codes starting with "V" from icd_query and check vocabulary_id using concept_id.
    final_query union distinct icd_query without V codes
    and v_icd_vocab_query which has V codes with proper vocabulary_ids.

    The reason for this is to ensure vocabulary_id values of V codes, many of which overlap between ICD9CM & ICD10CM,
    are correct.

    :param ds: Google BigQuery dataset ID containing OMOP data tables
    :return: a SQL query that would generate a table contains participant IDs and their ICD codes from unique dates
    """
    icd_query: str = f"""
        (
            SELECT DISTINCT
                co.person_id,
                co.condition_start_date AS date,
                c.vocabulary_id AS vocabulary_id,
                c.concept_code AS ICD,
                co.condition_concept_id AS concept_id
            FROM
                {ds}.condition_occurrence AS co
            INNER JOIN
                {ds}.concept AS c
            ON
                co.condition_source_value = c.concept_code
            WHERE
                c.vocabulary_id in ("ICD9CM", "ICD10CM")
        )
        UNION DISTINCT
        (
            SELECT DISTINCT
                co.person_id,
                co.condition_start_date AS date,
                c.vocabulary_id AS vocabulary_id,
                c.concept_code AS ICD,
                co.condition_concept_id AS concept_id
            FROM
                {ds}.condition_occurrence AS co
            INNER JOIN
                {ds}.concept AS c
            ON
                co.condition_source_concept_id = c.concept_id
            WHERE
                c.vocabulary_id in ("ICD9CM", "ICD10CM")
        )
        UNION DISTINCT
        (
            SELECT DISTINCT
                o.person_id,
                o.observation_date AS date,
                c.vocabulary_id AS vocabulary_id,
                c.concept_code AS ICD,
                o.observation_concept_id AS concept_id
            FROM
                {ds}.observation AS o
            INNER JOIN
                {ds}.concept as c
            ON
                o.observation_source_value = c.concept_code
            WHERE
                c.vocabulary_id in ("ICD9CM", "ICD10CM")
        )
        UNION DISTINCT
        (
            SELECT DISTINCT
                o.person_id,
                o.observation_date AS date,
                c.vocabulary_id AS vocabulary_id,
                c.concept_code AS ICD,
                o.observation_concept_id AS concept_id
            FROM
                {ds}.observation AS o
            INNER JOIN
                {ds}.concept as c
            ON
                o.observation_source_concept_id = c.concept_id
            WHERE
                c.vocabulary_id in ("ICD9CM", "ICD10CM")
        )
    """

    v_icd_vocab_query: str = f"""
        SELECT DISTINCT
            v_icds.person_id,
            v_icds.date,
            v_icds.ICD,
            c.vocabulary_id
        FROM
            (
                SELECT
                    *
                FROM
                    ({icd_query}) AS icd_events
                WHERE
                    icd_events.ICD LIKE "V%"
            ) AS v_icds
        INNER JOIN
            {ds}.concept_relationship AS cr
        ON
            v_icds.concept_id = cr.concept_id_1
        INNER JOIN
            {ds}. concept AS c
        ON
            cr.concept_id_2 = c.concept_id
        WHERE
            c.vocabulary_id IN ("ICD9CM", "ICD10CM")
        AND
            v_icds.ICD = c.concept_code
        AND NOT
            v_icds.vocabulary_id != c.vocabulary_id
    """

    final_query: str = f"""
        (
            SELECT DISTINCT
                person_id,
                date,
                ICD,
                vocabulary_id
            FROM 
                ({icd_query})
            WHERE
                NOT ICD LIKE "V%"
        )
        UNION DISTINCT
        (
            SELECT DISTINCT
                *
            FROM
                ({v_icd_vocab_query})
        )
    """

    return final_query


def natural_age_query(ds, participant_ids):
    """
    This method is exclusively for All of Us platform
    :param ds: Google BigQuery dataset ID containing OMOP data tables
    :param participant_ids: list of participant IDs to query
    :return: a SQL query that would generate a table
            contains participant IDs and their natural age
    """
    query: str = f"""
        SELECT
            DISTINCT p.person_id, 
            DATETIME_DIFF(
                IF(DATETIME(death_datetime) IS NULL, CURRENT_DATETIME(), DATETIME(death_datetime)), 
                DATETIME(birth_datetime), 
                DAY
            )/365.2425 AS natural_age
        FROM
            {ds}.person AS p
        LEFT JOIN
            {ds}.death AS d
        ON
            p.person_id = d.person_id
        WHERE
            p.person_id IN {participant_ids}
    """

    return query


def ehr_dx_code_query(ds, participant_ids):
    """
    This method is exclusively for All of Us platform.
    In condition occurrence table, diagnosis codes belongs to ICD9CM, ICD10CM, are counted.
    In observation table, diagnosis codes belongs to ICD9CM and ICD10CM are counted.
    :param ds: Google BigQuery dataset ID containing OMOP data tables
    :param participant_ids: list of participant IDs to query
    :return: a SQL query that would generate a table contains participant IDs and
            their ehr length (days), diagnosis code count(ICD), and age at last event
    """
    query: str = f"""
        SELECT DISTINCT
            df1.person_id,
            (DATETIME_DIFF(MAX(date), MIN(date), DAY) + 1)/365.2425 AS ehr_length,
            COUNT(code) AS dx_code_occurrence_count,
            COUNT(DISTINCT(code)) AS dx_condition_count,
            DATETIME_DIFF(MAX(date), MIN(birthday), DAY)/365.2425 AS age_at_last_event,
        FROM
            (
                (
                SELECT DISTINCT
                    co.person_id,
                    co.condition_start_date AS date,
                    c.concept_code AS code
                FROM
                    {ds}.condition_occurrence AS co
                INNER JOIN
                    {ds}.concept AS c
                ON
                    co.condition_source_value = c.concept_code
                WHERE
                    c.vocabulary_id IN ("ICD9CM", "ICD10CM")
                    AND
                    person_id IN {participant_ids}
                )
            UNION DISTINCT
                (
                SELECT DISTINCT
                    co.person_id,
                    co.condition_start_date AS date,
                    c.concept_code AS code
                FROM
                    {ds}.condition_occurrence AS co
                INNER JOIN
                    {ds}.concept AS c
                ON
                    co.condition_source_concept_id = c.concept_id
                WHERE
                    c.vocabulary_id IN ("ICD9CM", "ICD10CM")
                    AND
                    person_id IN {participant_ids}
                )
            UNION DISTINCT
                (
                SELECT DISTINCT
                    o.person_id,
                    o.observation_date AS date,
                    c.concept_code AS code
                FROM
                    {ds}.observation AS o
                INNER JOIN
                    {ds}.concept AS c
                ON
                    o.observation_source_value = c.concept_code
                WHERE
                    c.vocabulary_id IN ("ICD9CM", "ICD10CM")
                    AND
                    person_id IN {participant_ids}
                )
            UNION DISTINCT
                (
                SELECT DISTINCT
                    o.person_id,
                    o.observation_date AS date,
                    c.concept_code AS code
                FROM
                    {ds}.observation AS o
                INNER JOIN
                    {ds}.concept AS c
                ON
                    o.observation_source_concept_id = c.concept_id
                WHERE
                    c.vocabulary_id IN ("ICD9CM", "ICD10CM")
                    AND
                    person_id IN {participant_ids}
                )
            ) AS df1
        INNER JOIN
            (
                SELECT
                    person_id, 
                    EXTRACT(DATE FROM DATETIME(birth_datetime)) AS birthday
                FROM
                    {ds}.person
                WHERE
                    person_id IN {participant_ids}
            ) AS df2
        ON
            df1.person_id = df2.person_id
        GROUP BY 
            df1.person_id
    """

    return query


def sex_at_birth(ds, participant_ids):
    """
    This method is exclusively for All of Us platform
    :param ds: Google BigQuery dataset ID containing OMOP data tables
    :param participant_ids: list of participant IDs to query
    :return: a SQL query that would generate a table contains participant IDs and
            their sex at birth; male = 1, female = 0
    """
    query: str = f"""
        SELECT
            *
        FROM
            (
                (
                SELECT
                    person_id,
                    1 AS sex_at_birth
                FROM
                    {ds}.person
                WHERE
                    sex_at_birth_source_concept_id = 1585846
                AND
                    person_id IN {participant_ids}
                )
            UNION DISTINCT
                (
                SELECT
                    person_id,
                    0 AS sex_at_birth
                FROM
                    {ds}.person
                WHERE
                    sex_at_birth_source_concept_id = 1585847
                AND
                    person_id IN {participant_ids}
                )
            )
    """

    return query



# noinspection PyUnresolvedReferences,PyProtectedMember
from PheTK import _queries, _paths, _utils
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm
import os
import pandas as pd
import polars as pl
import sys


class Cohort:

    def __init__(self,
                 platform="aou",
                 aou_db_version=7,
                 aou_omop_cdr=None,
                 gbq_dataset_id=None):
        """
        :param platform: database; currently supports "aou" (All of Us) or "custom".
        :param aou_db_version: int type, version of database, e.g., 7 for All of Us CDR v7
        :param aou_omop_cdr: cdr string value, define where to query OMOP data;
                    if None, it will use current workspace CDR value, i.e., os.getenv("WORKSPACE_CDR")
        :param gbq_dataset_id: Google BigQuery dataset ID for custom platforms.
        """
        if platform.lower() != "aou" and platform.lower() != "custom":
            print("Unsupported database. Currently supports \"aou\" (All of Us) or \"custom\".")
            sys.exit(0)
        if platform.lower() == "aou" and aou_db_version != 6 and aou_db_version != 7:
            print("Unsupported database. Currently supports All of Us CDR v6 and v7 "
                  "(enter 6 or 7 as parameter value).")
            sys.exit(0)
        if platform.lower() == "custom" and gbq_dataset_id is None:
            print("custom_db is required for non All of Us platforms.")
            sys.exit(0)
        self.platform = platform.lower()

        # generate attributes for AoU class instance
        if self.platform == "aou":
            self.db_version = aou_db_version
            if aou_omop_cdr is None:
                self.cdr = os.getenv("WORKSPACE_CDR")
            else:
                self.cdr = aou_omop_cdr
            self.user_project = os.getenv("GOOGLE_PROJECT")
        else:
            self.cdr = gbq_dataset_id
                     
        # attributes for add_covariate method
        self.natural_age = False
        self.age_at_last_event = False
        self.sex_at_birth = False
        self.ehr_length = False
        self.dx_code_occurrence_count = False
        self.dx_condition_count = False
        self.genetic_ancestry = False
        self.first_n_pcs = 0

    def by_genotype(self,
                    chromosome_number,
                    genomic_position,
                    ref_allele,
                    alt_allele,
                    case_gt,
                    control_gt,
                    reference_genome="GRCh38",
                    mt_path=None,
                    output_file_name=None):
        """
        Generate cohort based on genotype of variant of interest
        :param chromosome_number: chromosome number; int
        :param genomic_position: genomic position; int
        :param ref_allele: reference allele; str
        :param alt_allele: alternative allele; str
        :param case_gt: genotype(s) for case; str or list of str
        :param control_gt: genotype(s) for control; str or list of str
        :param reference_genome: defaults to "GRCh38"; accepts "GRCh37" or "GRCh38"
        :param mt_path: path to population level Hail variant matrix table
        :param output_file_name: name of csv file output
        :return: genotype cohort csv file as well as polars dataframe object
        """

        # import hail and assign hail_init attribute if needed
        import hail as hl

        # set database path
        if self.platform == "aou":
            if mt_path is None and self.db_version == 6:
                mt_path = _paths.cdr6_mt_path
            elif mt_path is None and self.db_version == 7:
                mt_path = _paths.cdr7_mt_path
        elif self.platform == "custom" and mt_path is None:
            print("For custom platform, mt_path must not be None.")
            sys.exit(0)

        # basic data processing
        if output_file_name is not None:
            if ".csv" in output_file_name:
                output_file_name = output_file_name.replace(".csv", "")
            output_file_name = f"{output_file_name}.csv"
        else:
            output_file_name = "aou_chr" + \
                               str(chromosome_number) + "_" + \
                               str(genomic_position) + "_" + \
                               str(ref_allele) + "_" + \
                               str(alt_allele) + \
                               ".csv"
        if isinstance(case_gt, str):
            case_gt = [case_gt]
        if isinstance(control_gt, str):
            control_gt = [control_gt]
        gt_list = case_gt + control_gt
        alleles = f"{ref_allele}:{alt_allele}"
        base_locus = f"{chromosome_number}:{genomic_position}"
        if reference_genome == "GRCh38":
            locus = "chr" + base_locus
        elif reference_genome == "GRCh37":
            locus = base_locus
        else:
            print("Invalid reference version. Allowed inputs are \"GRCh37\" or \"GRCh38\".")
            return
        variant_string = locus + ":" + alleles

        # initialize Hail
        try:
            hl.init(default_reference=reference_genome)
        except Exception as err:
            if "IllegalArgumentException" not in str(err):
                raise
            else:
                print("Hail Initialization skipped as Hail has already been initialized.")

        # hail variant struct
        variant = hl.parse_variant(variant_string, reference_genome=reference_genome)

        # load and filter matrix table
        mt = hl.read_matrix_table(mt_path)
        mt = mt.filter_rows(mt.locus == hl.Locus.parse(locus))
        if mt.count_rows() == 0:
            print()
            print(f"\033[1mLocus {locus} not found!")
            return
        elif mt.count_rows() >= 1:
            print()
            print(f"\033[1mLocus {locus} found!")
            mt.row.show()

        # split if multi-allelic site
        allele_count = _utils.spark_to_polars(mt.entries().select("info").to_spark())
        allele_count = len(allele_count["info.AF"][0])
        if allele_count > 1:
            print()
            print("\033[1mMulti-allelic detected! Splitting...")
            mt = hl.split_multi(mt)
            mt.row.show()

        # keep variant of interest
        mt = mt.filter_rows((mt.locus == variant["locus"]) &
                            (mt.alleles == variant["alleles"]))
        if mt.count_rows() >= 1:
            print()
            print(f"\033[1mVariant {variant_string} found!")
            mt.row.show()

            # export to polars
            spark_df = mt.entries().select("GT").to_spark()
            polars_df = _utils.spark_to_polars(spark_df)

            # convert list of int to GT string, e.g., "0/0", "0/1", "1/1"
            polars_df = polars_df.with_columns(
                pl.col("GT.alleles").list.get(0).cast(pl.Utf8).alias("GT0"),
                pl.col("GT.alleles").list.get(1).cast(pl.Utf8).alias("GT1"),
            )
            polars_df = polars_df.with_columns((pl.col("GT0") + "/" + pl.col("GT1")).alias("GT"))
            polars_df = polars_df.filter(pl.col("GT").is_in(gt_list))
            polars_df = polars_df.with_columns(pl.when(pl.col("GT").is_in(case_gt))
                                               .then(1)
                                               .otherwise(0)
                                               .alias("case"))
            cohort = polars_df \
                .rename({"s": "person_id"})[["person_id", "case"]] \
                .with_columns(pl.col("person_id").cast(int))
            cohort = cohort.unique()
            cohort.write_csv(output_file_name)

            print()
            print("\033[1mCohort size:", len(cohort))
            print("\033[1mCases:", cohort["case"].sum())
            print("\033[1mControls:", len(cohort.filter(pl.col("case") == 0)), "\033[0m")
            print()
            print(f"\033[1mCohort data saved as {output_file_name}!\033[0m")
            print()

        else:
            print()
            print(f"Variant {variant_string} not found!")
            print()

    def _get_ancestry_preds(self, user_project, participant_ids):
        """
        This method specifically designed for All of Us database
        :param user_project: proxy of GOOGLE_PROJECT environment variable of current workspace in All of Us workbench
        :param participant_ids: participant IDs of interest
        :return: ancestry_preds data of specific version as polars dataframe object
        """
        if self.db_version == 7:
            ancestry_preds = pd.read_csv(_paths.cdr7_ancestry_pred_path,
                                         sep="\t",
                                         storage_options={"requester_pays": True,
                                                          "user_project": user_project})
            ancestry_preds = pl.from_pandas(ancestry_preds)
            ancestry_preds = ancestry_preds.with_columns(pl.col("pca_features").str.replace(r"\[", "")) \
                .with_columns(pl.col("pca_features").str.replace(r"\]", "")) \
                .with_columns(pl.col("pca_features").str.split(",").list.get(i).alias(f"pc{i}") for i in range(16)) \
                .with_columns(pl.col(f"pc{i}").str.replace(" ", "").cast(float) for i in range(16)) \
                .drop(["probabilities", "pca_features", "ancestry_pred_other"]) \
                .rename({"research_id": "person_id",
                         "ancestry_pred": "genetic_ancestry"}) \
                .filter(pl.col("person_id").is_in(participant_ids))
        else:
            ancestry_preds = None

        return ancestry_preds

    def _get_covariates(self, participant_ids):
        """
        This method specifically designed for All of Us database
        Core internal function to generate covariate data for a set of participant IDs
        :param participant_ids: IDs of interest
        :return: polars dataframe object
        """

        # initial data prep
        if isinstance(participant_ids, str) or isinstance(participant_ids, int):
            participant_ids = (participant_ids,)
        elif isinstance(participant_ids, list):
            participant_ids = tuple(participant_ids)
        df = pl.DataFrame({"person_id": participant_ids})
        participant_ids = tuple([int(i) for i in participant_ids])

        # GET COVARIATES
        # natural_age
        if self.natural_age:
            natural_age_df = _utils.polars_gbq(_queries.natural_age_query(self.cdr, participant_ids))
            df = df.join(natural_age_df, how="left", on="person_id")

        # age_at_last_event, ehr_length, dx_code_occurrence_count, dx_condition_count
        if self.age_at_last_event or self.ehr_length or self.dx_code_occurrence_count or self.dx_condition_count:
            temp_df = _utils.polars_gbq(_queries.ehr_dx_code_query(self.cdr, participant_ids))
            cols_to_keep = ["person_id"]
            if self.age_at_last_event:
                cols_to_keep.append("age_at_last_event")
            if self.ehr_length:
                cols_to_keep.append("ehr_length")
            if self.dx_code_occurrence_count:
                cols_to_keep.append("dx_code_occurrence_count")
            if self.dx_condition_count:
                cols_to_keep.append("dx_condition_count")
            df = df.join(temp_df[cols_to_keep], how="left", on="person_id")

        # sex_at_birth
        if self.sex_at_birth:
            sex_df = _utils.polars_gbq(_queries.sex_at_birth(self.cdr, participant_ids))
            df = df.join(sex_df, how="left", on="person_id")

        # genetic_ancestry, first_n_pcs
        if self.genetic_ancestry or self.first_n_pcs > 0:
            temp_df = self._get_ancestry_preds(self.user_project, participant_ids)
            cols_to_keep = ["person_id"]
            if self.genetic_ancestry:
                cols_to_keep.append("genetic_ancestry")
            if self.first_n_pcs > 0:
                cols_to_keep = cols_to_keep + [f"pc{i}" for i in range(self.first_n_pcs)]
            df = df.join(temp_df[cols_to_keep], how="left", on="person_id")

        return df

    def add_covariates(self,
                       cohort_csv_path=None,
                       natural_age=False,
                       age_at_last_event=False,
                       sex_at_birth=True,
                       ehr_length=False,
                       dx_code_occurrence_count=False,
                       dx_condition_count=False,
                       genetic_ancestry=False,
                       first_n_pcs=0,
                       chunk_size=10000,
                       drop_nulls=False,
                       output_file_name=None):
        """
        This method is a proxy for covariate.get_covariates method
        :param cohort_csv_path:
        :param natural_age: age of participants as of today
        :param age_at_last_event: age of participants at their last diagnosis event in EHR record
        :param sex_at_birth: sex at birth from survey and observation
        :param ehr_length: number of days that EHR record spans
        :param dx_code_occurrence_count: count of diagnosis code occurrences on unique dates,
                                         including ICD9CM, ICD10CM & SNOMED, throughout participant EHR history
        :param dx_condition_count: count of unique condition (dx code) throughout participant EHR history
        :param genetic_ancestry: predicted ancestry based on sequencing data
        :param first_n_pcs: number of first principal components to include
        :param chunk_size: defaults to 10,000; number of IDs per thread
        :param drop_nulls: defaults to False; drop rows having null values, i.e., participants without all covariates
        :param output_file_name: name for output csv file; do not include ".csv"
        :return: csv file and polars dataframe object
        """
        # assign attributes
        self.natural_age = natural_age
        self.age_at_last_event = age_at_last_event
        self.sex_at_birth = sex_at_birth
        self.ehr_length = ehr_length
        self.dx_code_occurrence_count = dx_code_occurrence_count
        self.dx_condition_count = dx_condition_count
        self.genetic_ancestry = genetic_ancestry
        self.first_n_pcs = first_n_pcs

        # check for valid input
        if cohort_csv_path is not None:
            cohort = pl.read_csv(cohort_csv_path)
            if "person_id" not in cohort.columns:
                print("Cohort must contains \"person_id\" column!")
                sys.exit(0)
        else:
            print("A cohort is required."
                  "Please provide a valid file path.")
            sys.exit(0)

        # get participant IDs from cohort
        participant_ids = cohort["person_id"].unique().to_list()

        # setup multi-threading to generate covariates
        chunks = [
            list(participant_ids)[i * chunk_size:(i + 1) * chunk_size] for i in
            range((len(participant_ids) // chunk_size) + 1)
        ]
        with ThreadPoolExecutor() as executor:
            jobs = [
                executor.submit(
                    self._get_covariates,
                    chunk
                ) for chunk in chunks
            ]
            result_list = [job.result() for job in tqdm(as_completed(jobs), total=len(chunks))]

        # process result
        result_list = [result for result in result_list if result is not None]
        covariates = result_list[0]
        for i in range(1, len(chunks)):
            covariates = pl.concat([covariates, result_list[i]])

        covariates = covariates.unique()
        covariates.write_csv("covariates.csv")

        # merge covariates to cohort
        final_cohort = cohort.join(covariates, how="left", on="person_id")
        if drop_nulls:
            final_cohort = final_cohort.drop_nulls()

        file_name = "cohort"
        if output_file_name is not None:
            if ".csv" in output_file_name:
                output_file_name = output_file_name.replace(".csv", "")
            file_name = output_file_name
        final_cohort.write_csv(f"{file_name}.csv")

        print()
        print("Cohort size:", len(final_cohort))
        if "case" in final_cohort.columns:
            print("Cases:", final_cohort["case"].sum())
            print("Controls:", len(final_cohort.filter(pl.col("case") == 0)), "\033[0m")
        print()
        print(f"Cohort data saved as \"{file_name}.csv\"!\033[0m")
        print()



from google.cloud import bigquery
import os
import polars as pl
import pyarrow as pa
import sys


def to_polars(df):
    """
    Check and convert pandas dataframe object to polars dataframe, if applicable
    :param df: dataframe object
    :return: polars dataframe
    """
    if not isinstance(df, pl.DataFrame):
        return pl.from_pandas(df)
    else:
        return df


def spark_to_polars(spark_df):
    """
    Convert spark df to polars df
    :param spark_df: spark df
    :return: polars df
    """
    # noinspection PyProtectedMember,PyArgumentList
    polars_df = pl.from_arrow(pa.Table.from_batches(spark_df._collect_as_arrow()))

    return polars_df


def polars_gbq(query):
    """
    Take a SQL query and return result as polars dataframe
    :param query: BigQuery SQL query
    :return: polars dataframe
    """
    client = bigquery.Client()
    query_job = client.query(query)
    rows = query_job.result()
    df = pl.from_arrow(rows.to_arrow())

    return df


def get_phecode_mapping_table(phecode_version, icd_version, phecode_map_file_path, keep_all_columns=True):
    """
    Load phecode mapping table
    :param phecode_version: defaults to "X"; other option is "1.2"
    :param icd_version: defaults to "US"; other option are "WHO" and "custom";
                        if "custom", user need to provide phecode_map_path
    :param phecode_map_file_path: path to custom phecode map table
    :param keep_all_columns: defaults to True
    :return: phecode mapping table as polars dataframe
    """
    # load phecode mapping file by version or by custom path
    phetk_dir = os.path.dirname(__file__)
    final_file_path = os.path.join(phetk_dir, "phecode")
    path_suffix = ""
    if phecode_version == "X":
        if icd_version == "US":
            path_suffix = "phecodeX.csv"
        elif icd_version == "WHO":
            path_suffix = "phecodeX_WHO.csv"
        elif icd_version == "custom":
            if phecode_map_file_path is None:
                print("Please provide phecode_map_path for custom icd_version")
                sys.exit(0)
        else:
            print("Invalid icd_version. Available icd_version values are US, WHO and custom.")
            sys.exit(0)
        if phecode_map_file_path is None:
            final_file_path = os.path.join(final_file_path, path_suffix)
        else:
            final_file_path = phecode_map_file_path
        # noinspection PyTypeChecker
        phecode_df = pl.read_csv(final_file_path,
                                 dtypes={"phecode": str,
                                         "ICD": str,
                                         "flag": pl.Int8,
                                         "code_val": float})
        if not keep_all_columns:
            phecode_df = phecode_df[["phecode", "ICD", "flag"]]
    elif phecode_version == "1.2":
        if icd_version == "US":
            path_suffix = "phecode12.csv"
        elif icd_version == "WHO":
            print("PheTK does not support mapping ICD-10 (WHO version) to phecode 1.2")
            sys.exit(0)
        elif icd_version == "custom":
            if phecode_map_file_path is None:
                print("Please provide phecode_map_path for custom icd_version")
                sys.exit(0)
        else:
            print("Invalid icd_version. Available icd_version values are US, WHO and custom.")
            sys.exit(0)
        if phecode_map_file_path is None:
            final_file_path = os.path.join(final_file_path, path_suffix)
        else:
            final_file_path = phecode_map_file_path
        # noinspection PyTypeChecker
        phecode_df = pl.read_csv(final_file_path,
                                 dtypes={"phecode": str,
                                         "ICD": str,
                                         "flag": pl.Int8,
                                         "exclude_range": str,
                                         "phecode_unrolled": str})
        if not keep_all_columns:
            phecode_df = phecode_df[["phecode_unrolled", "ICD", "flag"]]
    else:
        print("Unsupported phecode version. Supports phecode \"1.2\" and \"X\".")
        sys.exit(0)

    return phecode_df



from datetime import datetime
from matplotlib.lines import Line2D
import adjustText
import matplotlib.colors as mc
import matplotlib.pyplot as plt
import numpy as np
import polars as pl


class Plot:
    def __init__(self,
                 phewas_result_csv_path,
                 bonferroni=None,
                 phecode_version=None,
                 color_palette=None):
        """
        :param phewas_result_csv_path: path to PheWAS result csv file, generated from PheWAS module.
        :param bonferroni: defaults to None; if None, calculate base on number of phecode tested
        :param phecode_version: defaults to None; if None, use phecode X; else phecode 1.2
        :param color_palette: defaults to None; if None, use internal color palette
        """

        # load PheWAS results
        self.phewas_result = pl.read_csv(phewas_result_csv_path, dtypes={"phecode": str})

        # assign a proxy value for infinity neg_log_p_value
        max_non_inf_neg_log = self.phewas_result.filter(pl.col("neg_log_p_value") != np.inf) \
            .sort(by="neg_log_p_value", descending=True)["neg_log_p_value"][0]
        if max_non_inf_neg_log < self.phewas_result["neg_log_p_value"].max():
            self.inf_proxy = max_non_inf_neg_log * 1.2
            self.phewas_result = self.phewas_result.with_columns(pl.when(pl.col("neg_log_p_value") == np.inf)
                                                                 .then(self.inf_proxy)
                                                                 .otherwise(pl.col("neg_log_p_value"))
                                                                 .alias("neg_log_p_value"))
        else:
            self.inf_proxy = None

        # bonferroni
        if not bonferroni:
            self.bonferroni = -np.log10(0.05 / len(self.phewas_result))
        else:
            self.bonferroni = bonferroni

        # nominal significance
        self.nominal_significance = -np.log10(0.05)

        # phecode_version
        if phecode_version:
            self.phecode_version = phecode_version.upper()
        else:
            self.phecode_version = "X"

        # phecode categories
        self.phecode_categories = None

        # color mapping
        if color_palette:
            self.color_palette = color_palette
        else:
            self.color_palette = ("blue", "indianred", "darkcyan", "goldenrod", "darkblue", "magenta",
                                  "green", "red", "darkturquoise", "olive", "black", "royalblue",
                                  "maroon", "darkolivegreen", "coral", "purple", "gray")
        self.phecode_categories = self.phewas_result["phecode_category"].unique().to_list()
        self.phecode_categories.sort()
        self.color_dict = {self.phecode_categories[i]: self.color_palette[i % len(self.color_palette)]
                           for i in range(len(self.phecode_categories))}
        self.phewas_result = self.phewas_result.with_columns(
            pl.col("phecode_category").map_dict(self.color_dict).alias("label_color")
        )
        self.positive_betas = None
        self.negative_betas = None
        self.data_to_label = None

        # plot element scaling ratio
        self.ratio = 1

        # color lightness
        self.positive_alpha = 0.7
        self.negative_alpha = 0.3

        # offset
        self.offset = 9

        # volcano plot label data
        self.volcano_label_data = None

    @staticmethod
    def save_plot(plot_type="plot", output_file_name=None, output_file_type="pdf"):
        if output_file_name is not None:
            if "." not in output_file_name:
                output_file_name = output_file_name + "." + output_file_type
        else:
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            output_file_name = f"{plot_type}_{timestamp}.{output_file_type}"
        plt.savefig(output_file_name, bbox_inches="tight")
        print()
        print("Plot saved to", output_file_name)
        print()

    @staticmethod
    def _filter_by_phecode_categories(df, phecode_categories=None):
        """
        :param df: PheWAS result to filter
        :param phecode_categories: defaults to None, i.e., use all, otherwise, filter what specified
        :return: filtered df
        """
        if phecode_categories:
            if isinstance(phecode_categories, str):
                phecode_categories = [phecode_categories]
            df = df.filter(pl.col("phecode_category").is_in(phecode_categories))
        else:
            df = df

        return df

    @staticmethod
    def _create_phecode_index(df):
        """
        Create phecode index after grouping by phecode_category and phecode;
        Phecode index will be used for plotting purpose
        :param df: PheWAS result to create index
        :return: same dataframe with column "phecode_index" created
        """
        if "phecode_index" in df.columns:
            df = df.drop("phecode_index")
        df = df.sort(by=["phecode_category", "phecode"])\
               .with_columns(pl.Series("phecode_index", range(1, len(df) + 1)))\
               .with_columns(15*np.exp(pl.col("beta")).alias("marker_size_by_beta"))

        return df

    @staticmethod
    def _split_by_beta(df, marker_size_by_beta=False):
        """
        :param df: data of interest, e.g., full phewas result or result of a phecode_category
        :return: positive and negative beta polars dataframes
        """

        # add marker size if marker_size_by_beta is True
        if marker_size_by_beta:
            df = df.with_columns((18*pl.col("beta").abs()).alias("_marker_size"))

        # split to positive and negative beta data
        positive_betas = df.filter(pl.col("beta") >= 0).sort(by="beta", descending=True)
        negative_betas = df.filter(pl.col("beta") < 0).sort(by="beta", descending=False)
        return positive_betas, negative_betas

    @staticmethod
    def _x_ticks(plot_df, selected_color_dict, size=8):
        """
        Generate x tick labels and colors
        :param plot_df: plot data
        :param selected_color_dict: color dict; this is changed based on number of phecode categories selected
        :return: x tick labels and colors for the plot
        """
        x_ticks = plot_df[["phecode_category", "phecode_index"]].groupby("phecode_category").mean()
        # create x ticks labels and colors
        adjustText.plt.xticks(x_ticks["phecode_index"],
                              x_ticks["phecode_category"],
                              rotation=45,
                              ha="right",
                              weight="normal",
                              size=size)
        tick_labels = adjustText.plt.gca().get_xticklabels()
        sorted_labels = sorted(tick_labels, key=lambda label: label.get_text())
        for tick_label, tick_color in zip(sorted_labels, selected_color_dict.values()):
            tick_label.set_color(tick_color)

    def _manhattan_scatter(self, ax, marker_size_by_beta):
        """
        Generate scatter data points
        :param ax: plot object
        :param marker_size_by_beta: adjust marker size by beta coefficient if True
        :return: scatter plot of selected data
        """

        if marker_size_by_beta:
            s_positive = self.positive_betas["_marker_size"]
            s_negative = self.negative_betas["_marker_size"]
        else:
            s_positive = None
            s_negative = None

        ax.scatter(x=self.positive_betas["phecode_index"].to_numpy(),
                   y=self.positive_betas["neg_log_p_value"],
                   s=s_positive,
                   c=self.positive_betas["label_color"],
                   marker="^",
                   alpha=self.positive_alpha)

        ax.scatter(x=self.negative_betas["phecode_index"].to_numpy(),
                   y=self.negative_betas["neg_log_p_value"],
                   s=s_negative,
                   c=self.negative_betas["label_color"],
                   marker="v",
                   alpha=self.negative_alpha)

    def _lines(self,
               ax,
               plot_type,
               plot_df,
               x_col,
               nominal_significance_line=False,
               bonferroni_line=False,
               infinity_line=False,
               y_threshold_line=False,
               y_threshold_value=None,
               x_positive_threshold_line=False,
               x_positive_threshold_value=None,
               x_negative_threshold_line=False,
               x_negative_threshold_value=None):

        extra_offset = 0
        if plot_type == "manhattan":
            extra_offset = 1
        elif plot_type == "volcano":
            extra_offset = 0.05

        # nominal significance line
        if nominal_significance_line:
            ax.hlines(y=-adjustText.np.log10(.05),
                      xmin=plot_df[x_col].min() - self.offset - extra_offset,
                      xmax=plot_df[x_col].max() + self.offset + extra_offset,
                      colors="red",
                      lw=1)

        # bonferroni
        if bonferroni_line:
            ax.hlines(y=self.bonferroni,
                      xmin=plot_df[x_col].min() - self.offset - extra_offset,
                      xmax=plot_df[x_col].max() + self.offset + extra_offset,
                      colors="green",
                      lw=1)

        # infinity
        if infinity_line:
            if self.inf_proxy is not None:
                ax.yaxis.get_major_ticks()[-2].set_visible(False)
                ax.hlines(y=self.inf_proxy * 0.98,
                          xmin=plot_df[x_col].min() - self.offset - extra_offset,
                          xmax=plot_df[x_col].max() + self.offset + extra_offset,
                          colors="blue",
                          linestyle="dashdot",
                          lw=1)

        # y threshold line
        if y_threshold_line:
            ax.hlines(y=y_threshold_value,
                      xmin=plot_df[x_col].min() - self.offset - extra_offset,
                      xmax=plot_df[x_col].max() + self.offset + extra_offset,
                      colors="gray",
                      linestyles="dashed",
                      lw=1)

        # vertical lines
        if x_positive_threshold_line:
            ax.vlines(x=x_positive_threshold_value,
                      ymin=plot_df["neg_log_p_value"].min()-self.offset,
                      ymax=plot_df["neg_log_p_value"].max() + self.offset + 5,
                      colors="orange",
                      linestyles="dashed",
                      lw=1)
        if x_negative_threshold_line:
            ax.vlines(x=x_negative_threshold_value,
                      ymin=plot_df["neg_log_p_value"].min()-self.offset,
                      ymax=plot_df["neg_log_p_value"].max() + self.offset + 5,
                      colors="lightseagreen",
                      linestyles="dashed",
                      lw=1)

    @staticmethod
    def _split_text(s, threshold=30):
        """
        Split long text label
        :param s: text string
        :param threshold: approximate number of characters per line
        :return: split text if longer than 40 chars
        """
        words = s.split(" ")
        new_s = ""
        line_length = 0
        for word in words:
            new_s += word
            line_length += len(word)
            if line_length >= threshold and word != words[-1]:
                new_s += "\n"
                line_length = 0
            else:
                new_s += " "

        return new_s

    def _manhattan_label(self,
                         plot_df,
                         label_values,
                         label_count,
                         label_categories=None,
                         label_text_column="phecode_string",
                         label_value_threshold=0,
                         label_split_threshold=30,
                         label_color="label_color",
                         label_size=8,
                         label_weight="normal",
                         y_col="neg_log_p_value",
                         x_col="phecode_index"):
        """
        :param plot_df: plot data
        :param label_values: can take a single phecode, a list of phecodes,
                             or preset values "positive_betas", "negative_betas", "p_value"
        :param label_value_threshold: cutoff value for label values;
                                      if label_values is "positive_beta", keep beta values >= cutoff
                                      if label_values is "negative_beta", keep beta values <= cutoff
                                      if label_values is "p_value", keep neg_log_p_value >= cutoff
        :param label_text_column: defaults to "phecode_string"; name of column contain text for labels
        :param label_count: number of items to label, only needed if label_by input is data type
        :param label_split_threshold: number of characters to consider splitting long labels
        :param label_color: string type; takes either a color or name of column contains color for plot data
        :param label_size: defaults to 8
        :param label_weight: takes standard plt weight inputs, e.g., "normal", "bold", etc.
        :param x_col: column contains x values
        :param y_col: column contains y values
        :return: adjustText object
        """

        if isinstance(label_values, str):
            label_values = [label_values]

        self.data_to_label = pl.DataFrame(schema=plot_df.schema)
        positive_betas = self.positive_betas.clone()
        negative_betas = self.negative_betas.clone()
        if "_marker_size" in positive_betas.columns:
            positive_betas = positive_betas.drop("_marker_size")
        if "_marker_size" in negative_betas.columns:
            negative_betas = negative_betas.drop("_marker_size")

        for item in label_values:
            if item == "positive_beta":
                self.data_to_label = pl.concat(
                    [
                        self.data_to_label,
                        positive_betas.filter(pl.col("beta") >= label_value_threshold)
                    ]
                )
                if label_categories is not None:
                    self.data_to_label = self.data_to_label.filter(
                        pl.col("phecode_category").is_in(label_categories)
                    )[:label_count]
                else:
                    self.data_to_label = self.data_to_label[:label_count]
            elif item == "negative_beta":
                self.data_to_label = pl.concat(
                    [
                        self.data_to_label,
                        negative_betas.filter(pl.col("beta") <= label_value_threshold)
                    ]
                )
                if label_categories is not None:
                    self.data_to_label = self.data_to_label.filter(
                        pl.col("phecode_category").is_in(label_categories)
                    )[:label_count]
                else:
                    self.data_to_label = self.data_to_label[:label_count]
            elif item == "p_value":
                self.data_to_label = pl.concat(
                    [
                        self.data_to_label,
                        plot_df.sort(by="p_value")
                               .filter(pl.col("neg_log_p_value") >= label_value_threshold)
                    ]
                )
                if label_categories is not None:
                    self.data_to_label = self.data_to_label.filter(
                        pl.col("phecode_category").is_in(label_categories)
                    )[:label_count]
                else:
                    self.data_to_label = self.data_to_label[:label_count]
            else:
                self.data_to_label = pl.concat([self.data_to_label,
                                                plot_df.filter(pl.col("phecode") == item)])

        texts = []
        for i in range(len(self.data_to_label)):
            if mc.is_color_like(label_color):
                color = pl.Series(values=[label_color]*len(self.data_to_label))
            else:
                # noinspection PyTypeChecker
                color = self.data_to_label[label_color]
            # noinspection PyTypeChecker
            texts.append(adjustText.plt.text(float(self.data_to_label[x_col][i]),
                                             float(self.data_to_label[y_col][i]),
                                             self._split_text(self.data_to_label[label_text_column][i],
                                                              label_split_threshold),
                                             color=color[i],
                                             size=label_size,
                                             weight=label_weight,
                                             alpha=1,
                                             bbox=dict(facecolor="white",
                                                       edgecolor="none",
                                                       boxstyle="round",
                                                       alpha=0.5,
                                                       lw=0.5)))

        if len(texts) > 0:
            return adjustText.adjust_text(texts,
                                          arrowprops=dict(arrowstyle="simple", color="gray", lw=0.5, mutation_scale=2))

    def _manhattan_legend(self, ax, legend_marker_size):
        """
        :param ax: plot object
        :param legend_marker_size: size of markers
        :return: legend element
        """
        legend_elements = [
            Line2D([0], [0], color="blue", lw=1, linestyle="dashdot", label="Infinity"),
            Line2D([0], [0], color="green", lw=1, label="Bonferroni\nCorrection"),
            Line2D([0], [0], color="red", lw=1, label="Nominal\nSignificance"),
            Line2D([0], [0], marker="^", label="Increased\nRisk Effect", color="white",
                   markerfacecolor="blue", alpha=self.positive_alpha, markersize=legend_marker_size),
            Line2D([0], [0], marker="v", label="Decreased\nRisk Effect", color="white",
                   markerfacecolor="blue", alpha=self.negative_alpha, markersize=legend_marker_size),
        ]
        ax.legend(handles=legend_elements,
                  handlelength=2,
                  loc="center left",
                  bbox_to_anchor=(1, 0.5),
                  fontsize=legend_marker_size)

    def manhattan(self,
                  label_values="p_value",
                  label_value_threshold=0,
                  label_count=10,
                  label_size=8,
                  label_text_column="phecode_string",
                  label_color="label_color",
                  label_weight="normal",
                  label_split_threshold=30,
                  marker_size_by_beta=False,
                  phecode_categories=None,
                  plot_all_categories=True,
                  title=None,
                  title_text_size=10,
                  y_limit=None,
                  axis_text_size=8,
                  show_legend=True,
                  legend_marker_size=6,
                  dpi=150,
                  save_plot=True,
                  output_file_name=None,
                  output_file_type="pdf"):

        ############
        # SETTINGS #
        ############

        # setup some variables based on plot_all_categories and phecode_categories

        # offset
        self.offset = 9
        
        # phecode_categories & label_categories
        if phecode_categories:
            if isinstance(phecode_categories, str):  # convert to list if input is str
                phecode_categories = [phecode_categories]
            phecode_categories.sort()
            label_categories = phecode_categories
            self.phecode_categories = phecode_categories
        else:
            label_categories = None
        
        # plot_df and label_value_cols
        if plot_all_categories:
            selected_color_dict = self.color_dict
            n_categories = len(self.phewas_result.columns)
            # create plot_df containing only necessary data for plotting
            plot_df = self._create_phecode_index(self.phewas_result)
        else:
            if phecode_categories:
                selected_color_dict = {k: self.color_dict[k] for k in phecode_categories}
                n_categories = len(phecode_categories)
                dpi = None
                # create plot_df containing only necessary data for plotting
                plot_df = self._create_phecode_index(
                    self._filter_by_phecode_categories(
                        self.phewas_result, phecode_categories=phecode_categories
                    )
                )
            else:
                print("phecode_categories must not be None when plot_all_categories = False.")
                return

        # create plot
        self.ratio = (n_categories/len(self.phewas_result.columns))
        fig, ax = adjustText.plt.subplots(figsize=(12*self.ratio, 7), dpi=dpi)

        # plot title
        if title is not None:
            adjustText.plt.title(title, weight="bold", size=title_text_size)

        # set limit for display on y axes
        if y_limit is not None:
            ax.set_ylim(-0.2, y_limit)

        # y axis label
        ax.set_ylabel(r"$-\log_{10}$(p-value)", size=axis_text_size)

        # generate positive & negative betas
        self.positive_betas, self.negative_betas = self._split_by_beta(plot_df, marker_size_by_beta)

        ############
        # PLOTTING #
        ############

        # x-axis offset
        adjustText.plt.xlim(float(plot_df["phecode_index"].min()) - self.offset - 1,
                            float(plot_df["phecode_index"].max()) + self.offset + 1)

        # create x ticks labels and colors
        self._x_ticks(plot_df, selected_color_dict)

        # scatter
        self._manhattan_scatter(ax=ax, marker_size_by_beta=marker_size_by_beta)

        # lines
        self._lines(ax=ax,
                    plot_type="manhattan",
                    plot_df=plot_df,
                    x_col="phecode_index",
                    nominal_significance_line=True,
                    bonferroni_line=True,
                    infinity_line=True)

        # labeling
        self._manhattan_label(plot_df=plot_df, label_values=label_values, label_count=label_count,
                              label_text_column=label_text_column, label_categories=label_categories,
                              label_value_threshold=label_value_threshold, label_split_threshold=label_split_threshold,
                              label_size=label_size, label_color=label_color, label_weight=label_weight)

        # legend
        if show_legend:
            self._manhattan_legend(ax, legend_marker_size)

        # save plot
        if save_plot:
            self.save_plot(plot_type="manhattan",
                           output_file_name=output_file_name,
                           output_file_type=output_file_type)

    @staticmethod
    def transform_values(df, col, new_col, new_min, new_max):
        df = df.with_columns(((pl.col(col) - pl.col(col).min())
                              * (new_max - new_min)
                              / (pl.col(col).max() - pl.col(col).min())
                              + new_min).alias(new_col))
        return df

    def _volcano_scatter(self,
                         ax,
                         x_col="log10_odds_ratio",
                         y_col="neg_log_p_value",
                         marker_size_col="cases",
                         marker_shape=".",
                         positive_beta_color="indianred",
                         negative_beta_color="darkcyan",
                         fill_marker=True,
                         marker_alpha=0.5,
                         legend_marker_scale=0.5,
                         legend_label_count=5,
                         show_legend=False):

        # set marker edge and face colors
        if fill_marker:
            positive_face_color = positive_beta_color
            negative_face_color = negative_beta_color
        else:
            positive_face_color = "none"
            negative_face_color = "none"

        # color values for every points
        if marker_size_col is not None:
            col_list = [x_col, y_col, marker_size_col]
        else:
            col_list = [x_col, y_col]
        pos_df = self.positive_betas[col_list].with_columns(pl.lit(positive_beta_color)
                                                            .alias("edge_color"))\
                                              .with_columns(pl.lit(positive_face_color)
                                                            .alias("face_color"))
        neg_df = self.negative_betas[col_list].with_columns(pl.lit(negative_beta_color)
                                                            .alias("edge_color"))\
                                              .with_columns(pl.lit(negative_face_color)
                                                            .alias("face_color"))
        # combined into 1 df for plotting
        full_df = pl.concat([pos_df, neg_df]).unique()
        if marker_size_col is not None:
            # scale values for better visualization
            full_df = self.transform_values(df=full_df,
                                            col=marker_size_col,
                                            new_col="_marker_size",
                                            new_min=50,
                                            new_max=1000)
            marker_size = full_df["_marker_size"].to_numpy()
        else:
            marker_size = None

        # plot scatter
        scatter = ax.scatter(
            x=full_df[x_col].to_numpy(),
            y=full_df[y_col],
            s=marker_size,
            edgecolors=full_df["edge_color"],
            facecolors=full_df["face_color"],
            marker=marker_shape,
            alpha=marker_alpha,
        )

        # legend
        k = 0.05
        min_size = full_df[marker_size_col].min()
        max_size = full_df[marker_size_col].max()
        margin = (max_size - min_size) * k
        step_size = (max_size - min_size) * (1-(2*k)) / (legend_label_count - 1)
        legend_labels = [
            min_size + margin + (i * step_size) for i in range(legend_label_count)
        ]
        legend_labels = [round(i, -2) for i in legend_labels]
        if (marker_size_col is not None) and show_legend:
            handles, labels = scatter.legend_elements(prop="sizes", alpha=0.5, num=legend_label_count)
            ax.legend(
                handles=handles,
                labels=legend_labels,  # override with original values
                markerscale=legend_marker_scale,
                loc="center left",
                bbox_to_anchor=(1, 0.5),
                title=marker_size_col
            )

    def _volcano_label(self,
                       plot_df,
                       phecode_list=None,
                       phecode_string_list=None,
                       x_col="log10_odds_ratio",
                       y_col="neg_log_p_value",
                       label_count=10,
                       label_text_column="phecode_string",
                       label_split_threshold=30,
                       label_size=8,
                       label_weight="normal",
                       y_threshold=5,
                       x_positive_threshold=None,
                       x_negative_threshold=None):

        # get data for labeling, either use a list of phecodes/phecode names of choice, or use x & y thresholds
        if (phecode_list is not None) or (phecode_string_list is not None):
            if isinstance(phecode_string_list, str):
                phecode_string_list = [phecode_string_list]
            if isinstance(phecode_list, str):
                phecode_list = [phecode_list]
            data_to_label = plot_df.filter((pl.col("phecode").is_in(phecode_list)) |
                                           (pl.col("phecode_string").is_in(phecode_string_list)))
        elif (y_threshold is not None) or (x_negative_threshold is not None) or (x_positive_threshold is not None):
            if y_threshold is None:
                y_threshold = plot_df["neg_log_p_value"].min()
            if x_negative_threshold is None:
                x_negative_threshold = plot_df[x_col].max()
            if x_positive_threshold is None:
                x_positive_threshold = plot_df[x_col].min()
            data_to_label = plot_df.filter(
                ((pl.col(x_col) >= x_positive_threshold) | (pl.col(x_col) <= x_negative_threshold)) &
                (pl.col("neg_log_p_value") >= y_threshold)
            )
            data_to_label = pl.concat(
                [data_to_label.top_k(by=x_col, descending=True, k=round(label_count/2), nulls_last=True),
                 data_to_label.top_k(by=x_col, descending=False, k=round(label_count/2), nulls_last=True)]
            ).unique()
        else:
            data_to_label = pl.concat(
                [plot_df.top_k(by=x_col, descending=True, k=round(label_count/2), nulls_last=True),
                 plot_df.top_k(by=x_col, descending=False, k=round(label_count/2), nulls_last=True)]
            ).unique()

        # label data
        self.volcano_label_data = data_to_label

        texts = []
        for i in range(len(data_to_label)):
            if data_to_label[x_col][i] < 0:
                color = "green"
            else:
                color = "red"
            # noinspection PyTypeChecker
            texts.append(adjustText.plt.text(float(data_to_label[x_col][i]),
                                             float(data_to_label[y_col][i]),
                                             self._split_text(data_to_label[label_text_column][i],
                                                              label_split_threshold),
                                             color=color,
                                             size=label_size,
                                             weight=label_weight,
                                             bbox=dict(facecolor="white",
                                                       edgecolor="none",
                                                       boxstyle="round",
                                                       alpha=0.5,
                                                       lw=0.5),
                                             alpha=1))
        if len(texts) > 0:
            return adjustText.adjust_text(
                texts, arrowprops=dict(arrowstyle="simple", color="gray", lw=0.5, mutation_scale=2)
            )

    def volcano(self,
                phecode_list=None,
                phecode_string_list=None,
                label_count=10,
                x_col="log10_odds_ratio",
                y_col="neg_log_p_value",
                x_axis_label=None,
                exclude_infinity=False,
                y_threshold=None,
                x_negative_threshold=None,
                x_positive_threshold=None,
                bonferroni_line=False,
                nominal_significance_line=False,
                infinity_line=False,
                y_limit=None,
                title=None,
                title_text_size=None,
                axis_text_size=None,
                marker_size_col="cases",
                marker_shape=".",
                fill_marker=True,
                marker_alpha=0.5,
                show_legend=False,
                legend_marker_scale=0.5,
                legend_label_count=5,
                dpi=150,
                save_plot=True,
                output_file_name=None,
                output_file_type="pdf"):

        # set offset
        self.offset = 0.1

        # create plot
        fig, ax = adjustText.plt.subplots(figsize=(12 * self.ratio, 7), dpi=dpi)

        # plot title
        if title is not None:
            adjustText.plt.title(title, weight="bold", size=title_text_size)

        # set limit for display on y axes
        if y_limit is not None:
            ax.set_ylim(-0.2, y_limit)

        # x, y axis label
        if x_col == "log10_odds_ratio":
            x_axis_label = r"$\log_{10}$(OR)"
        elif (x_col != "log10_odds_ratio") and (x_axis_label is None):
            x_axis_label = x_col
        ax.set_xlabel(x_axis_label, size=axis_text_size)
        ax.set_ylabel(r"$-\log_{10}$(p-value)", size=axis_text_size)

        # plot_df
        plot_df = self.phewas_result.clone()
        if exclude_infinity:
            plot_df = plot_df.filter(pl.col("neg_log_p_value") != self.inf_proxy)
        # generate positive & negative betas
        self.positive_betas, self.negative_betas = self._split_by_beta(plot_df)

        ############
        # PLOTTING #
        ############

        # scatter
        self._volcano_scatter(ax=ax,
                              x_col=x_col,
                              marker_size_col=marker_size_col,
                              marker_shape=marker_shape,
                              fill_marker=fill_marker,
                              marker_alpha=marker_alpha,
                              legend_marker_scale=legend_marker_scale,
                              legend_label_count=legend_label_count,
                              show_legend=show_legend)

        # lines
        x_positive_threshold_line = False
        x_negative_threshold_line = False
        y_threshold_line = False
        if x_positive_threshold is not None:
            x_positive_threshold_line = True
        if x_negative_threshold is not None:
            x_negative_threshold_line = True
        if y_threshold is not None:
            y_threshold_line = True
        self._lines(ax=ax,
                    plot_type="volcano",
                    plot_df=plot_df,
                    x_col=x_col,
                    y_threshold_line=y_threshold_line,
                    y_threshold_value=y_threshold,
                    x_positive_threshold_line=x_positive_threshold_line,
                    x_positive_threshold_value=x_positive_threshold,
                    x_negative_threshold_line=x_negative_threshold_line,
                    x_negative_threshold_value=x_negative_threshold,
                    bonferroni_line=bonferroni_line,
                    nominal_significance_line=nominal_significance_line,
                    infinity_line=infinity_line)

        # labels
        self._volcano_label(phecode_list=phecode_list,
                            phecode_string_list=phecode_string_list,
                            plot_df=plot_df,
                            label_count=label_count,
                            x_col=x_col,
                            y_col=y_col,
                            y_threshold=y_threshold,
                            x_positive_threshold=x_positive_threshold,
                            x_negative_threshold=x_negative_threshold)

        # save plot
        if save_plot:
            self.save_plot(plot_type="volcano",
                           output_file_name=output_file_name,
                           output_file_type=output_file_type)



cdr6_ancestry_pred_path = "gs://fc-aou-datasets-controlled/v6/wgs/vcf/aux/ancestry/ancestry_preds.tsv"
cdr7_ancestry_pred_path = "gs://fc-aou-datasets-controlled/v7/wgs/short_read/snpindel/aux/ancestry/ancestry_preds.tsv"

cdr6_mt_path = "gs://fc-aou-datasets-controlled/v6/wgs/hail.mt"
cdr7_mt_path = "gs://fc-aou-datasets-controlled/v7/wgs/short_read/snpindel/acaf_threshold_v7.1/multiMT/hail.mt"

# obsolete paths
# cdr7_mt_path = "gs://fc-aou-datasets-controlled/v7/wgs/short_read/snpindel/acaf_threshold/multiMT/hail.mt"



from setuptools import setup, find_packages

# Read contents of the README.md file
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='PheTK',
    version='0.1.42',
    # version='0.2.1rc5',
    python_requires='>=3.7',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    package_data={
        '': ['*.*'],
        'PheTK': ['phecode/*'],
    },
    url='https://github.com/nhgritctran/PheTK',
    license='GPL-3.0',
    author='Tam Tran',
    author_email='PheTK@mail.nih.gov',
    description='The Phenotype Toolkit',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        "adjusttext",
        "connectorx",
        "google-cloud-bigquery",
        "hail",
        "lifelines",
        "lxml",
        "matplotlib",
        "numpy",
        "pandas",
        "polars<=0.20.26",
        "pyarrow",
        "statsmodels",
        "tqdm"
    ]
)



# noinspection PyUnresolvedReferences
from PheTK.PheWAS import PheWAS
from tqdm import tqdm
import numpy as np
import polars as pl
import os
import random
import sys


def generate_examples(phecode="GE_979.2", cohort_size=500, var_type="binary",
                      data_has_both_sexes=True):
    # load phecode mapping file to get all phecodes
    phetk_dir = os.path.dirname(__file__)
    phecode_mapping_file_path = os.path.join(phetk_dir, "phecode")
    phecode_mapping_file_path = os.path.join(phecode_mapping_file_path, "phecodeX.csv")
    # noinspection PyTypeChecker
    phecode_df = pl.read_csv(phecode_mapping_file_path,
                             dtypes={"phecode": str,
                                     "ICD": str,
                                     "flag": pl.Int8,
                                     "code_val": float})
    phecodes = phecode_df["phecode"].unique().to_list()
    phecodes.remove(phecode)  # exclude target phecode for background data

    # mock cohort
    if data_has_both_sexes:
        n_sex = 2
    else:
        n_sex = 1
    cols = {"person_id": np.array(range(1, cohort_size + 1)),
            "age": np.random.randint(18, 80, cohort_size),
            "sex": np.random.randint(0, n_sex, cohort_size),
            "pc1": np.random.uniform(-1, 1, cohort_size),
            "pc2": np.random.uniform(-1, 1, cohort_size),
            "pc3": np.random.uniform(-1, 1, cohort_size)}
    if var_type == "binary":
        cols["independent_variable_of_interest"] = np.random.randint(0, 2, cohort_size)
    elif var_type == "continuous":
        cols["independent_variable_of_interest"] = np.random.uniform(0.1, 10, cohort_size)

    cohort = pl.from_dict(cols)
    if var_type == "binary":
        case_ids = cohort.filter(pl.col("independent_variable_of_interest") == 1)["person_id"].unique().to_list()
        ctrl_ids = cohort.filter(pl.col("independent_variable_of_interest") == 0)["person_id"].unique().to_list()
    elif var_type == "continuous":
        var_mean = np.mean(cols.get("independent_variable_of_interest"))
        case_ids = cohort.filter(pl.col("independent_variable_of_interest") >= var_mean)["person_id"].unique().to_list()
        ctrl_ids = cohort.filter(pl.col("independent_variable_of_interest") < var_mean)["person_id"].unique().to_list()
    else:
        print("Error: var_type can only be \"binary\" or \"continuous\".")
        return

    # mock phecode_counts
    phecode_counts = None
    for i in tqdm(range(cohort_size)):
        if i == 0:
            cols = {}
            # noinspection PyTypeChecker
            ids = random.sample(
                case_ids, np.random.randint(round(len(case_ids) * 0.5), len(case_ids) * 0.8)
            ) + ctrl_ids[:round(len(ctrl_ids) / 10)]
            cols["person_id"] = np.array(ids)
            cols["phecode"] = np.array([phecode] * len(ids))
            cols["count"] = np.random.randint(1, 3, len(ids))
        else:
            cols = {}
            ids = random.sample(cohort["person_id"].to_list(),
                                np.random.randint(round(cohort_size * 0.5), round(cohort_size * 0.9)))
            cols["person_id"] = np.array(ids)
            cols["phecode"] = np.array([phecodes[np.random.randint(1, len(phecodes))]] * len(ids))
            cols["count"] = np.random.randint(1, 10, len(ids))
        df = pl.from_dict(cols)
        if phecode_counts is None:
            phecode_counts = df
        else:
            phecode_counts = pl.concat([phecode_counts, df])
    phecode_counts = phecode_counts.unique(["person_id", "phecode"]).sort(by="person_id")

    # save data
    cohort.write_csv("example_cohort.csv")
    phecode_counts.write_csv("example_phecode_counts.csv")
    print(
        "Generated data saved to \033[1m\"example_cohort.csv\"\033[0m",
        "and \033[1m\"example_phecode_counts.csv\"\033[0m"
    )


def _prompt():
    print()
    answer = input("Press enter to continue...")
    if answer.lower() == "quit":
        print()
        print("\033[1mGood luck!\033[0m")
        sys.exit(0)
    print()


def run(covariates_cols=("age", "sex", "pc1", "pc2", "pc3"),
        independent_variable_of_interest="independent_variable_of_interest",
        phecode_to_process="all",
        verbose=False):
    print("\033[1mHello and welcome to PheTK PheWAS demo.\033[0m")
    print("This is a quick demonstration to introduce a basic PheWAS analysis using mock data.",
          "It should take less than 1 minute running without pauses.",
          "For a detailed tutorial, please check out the included jupyter notebooks.",
          "Enter \"quit\" in any prompt to quit.")
    _prompt()
    print("\033[1mFirst, let's create some example data.\033[0m")
    print()
    print(f"We will create an example cohort with covariates {covariates_cols}.",
          "This data also contains our variable of interest which can be binary or continuous.")
    print()
    print("In addition, we will also create an example phenotype profile data for this cohort.",
          "This table contains all phecodes mapped from ICD codes from each person's EHR and their counts.")
    print()
    var_type = input("Which data type would you like the variable of interest to be? (binary/continuous) ")
    while (var_type.lower() != "binary") and (var_type.lower() != "continuous") and (var_type.lower() != "quit"):
        var_type = input("Please enter either binary or continuous:")
    if var_type.lower() == "quit":
        print()
        print("\033[1mGood luck!\033[0m")
    else:
        data_has_both_sexes = input("Would you like data to have both sexes? (yes/no) ")
        while (data_has_both_sexes.lower() != "yes") \
                and (data_has_both_sexes.lower() != "no") \
                and (data_has_both_sexes.lower() != "quit"):
            data_has_both_sexes = input("Please enter either yes or no:")
        if data_has_both_sexes.lower() == "quit":
            print()
            print("\033[1mGood luck!\033[0m")
        else:
            if data_has_both_sexes.lower() == "yes":
                data_has_both_sexes = True
            else:
                data_has_both_sexes = False
            generate_examples(var_type=var_type, data_has_both_sexes=data_has_both_sexes)
    _prompt()
    print("\033[1mWe created a cohort of 500 people and here is how the cohort data look like:\033[0m")
    print(pl.read_csv("example_cohort.csv").head())
    _prompt()
    print("\033[1mHere is how the phecode count data look like:\033[0m")
    print(pl.read_csv("example_phecode_counts.csv", dtypes={"phecode": str}).head())
    _prompt()
    print("\033[1mNow we are ready to run PheWAS!\033[0m")
    print()
    print("If run in command line interface, the analysis below can be run with the following command:")
    print("\033[1mpython3 -m PheTK.PheWAS --cohort_csv_path\033[0m example_cohort.csv",
          "\033[1m--phecode_count_csv_path\033[0m example_phecode_counts.csv",
          "\033[1m--phecode_version\033[0m X",
          "\033[1m--sex_at_birth_col\033[0m sex",
          "\033[1m--covariates\033[0m age sex pc1 pc2 pc3",
          "\033[1m--independent_variable_of_interest\033[0m independent_variable_of_interest",
          "\033[1m--min_case\033[0m 50",
          "\033[1m--min_phecode_count\033[0m 2",
          "\033[1m--output_file_name\033[0m example_phewas_results.csv")
    print()
    input("\033[1mPress enter to run PheWAS!\033[0m")
    print()
    if isinstance(covariates_cols, tuple):
        covariates_cols = list(covariates_cols)
    phewas = PheWAS(cohort_csv_path="example_cohort.csv",
                    phecode_count_csv_path="example_phecode_counts.csv",
                    phecode_version="X",
                    sex_at_birth_col="sex",
                    phecode_to_process=phecode_to_process,
                    covariate_cols=list(covariates_cols),
                    independent_variable_of_interest=independent_variable_of_interest,
                    min_cases=50,
                    min_phecode_count=2,
                    output_file_name="example_phewas_results.csv",
                    verbose=verbose)
    phewas.run()
    print("\033[1mHere is how example_phewas_results.csv look like:\033[0m")
    if independent_variable_of_interest == "independent_variable_of_interest" and phecode_to_process == "all":
        print("In this example, we intentionally generated data with Cystic Fibrosis as a significant hit.")
    print(pl.read_csv("example_phewas_results.csv", dtypes={"phecode": str}).sort(by="p_value").head())
    print()
    print("\033[1mThis is the end of the demo!\033[0m")
    print()
    print("\033[1mGood luck!\033[0m")


if __name__ == "__main__":
    run()
